# coding=utf8
# Copyright 2020 Cognicept Systems
# Author: Swarooph Seshadri (swarooph@cognicept.systems)
# --> Pusher class handles pushing stuff to the Cognicept cloud

from dotenv import dotenv_values
from pathlib import Path
import os
import sys
import time
import glob
import requests
import datetime
import threading
import ntpath
import logging
import boto3
from boto3.s3.transfer import TransferConfig
from botocore.exceptions import ClientError
from cogniceptshell.common import bcolors


class ProgressPercentage(object):
    """
    Track upload progress for chunks uploaded to S3
    ...

    Parameters
    ----------
    TODO

    Methods
    -------
    TODO
    """

    def __init__(self, filename):
        self._filename = filename
        self._size = float(os.path.getsize(filename))
        self._seen_so_far = 0
        self._lock = threading.Lock()

    def __call__(self, bytes_amount):
        # To simplify, assume this is hooked up to a single filename
        with self._lock:
            self._seen_so_far += bytes_amount
            percentage = (self._seen_so_far / self._size) * 100
            sys.stdout.write(
                "\r%s  %s / %s  (%.2f%%)" % (
                    self._filename, self._seen_so_far, self._size,
                    percentage))
            sys.stdout.flush()


class Pusher:
    """
    TODO
    ...

    Parameters
    ----------
    TODO

    Methods
    -------
    TODO
    """
    def __init__(self):
        self.bucket_name = 'cognicept-bagfiles'

    def push(self, args):
        """
        Main push entry point
        """
        if args.bag:
            # Construct client
            self.construct_client(args.config)
            # Push bag
            self.push_bag(args)
        else:
            # start or stop must be provided
            print(
                bcolors.FAIL + "Required command is missing." + bcolors.ENDC)

    def construct_client(self, config_obj):
        """
        Utility function to create S3 client based on runtime.env credentials
        """
        local_cfg = config_obj.config

        if 'AWS_SESSION_TOKEN' in local_cfg:
            self.s3_client = boto3.client(
                's3', region_name='ap-southeast-1',
                aws_access_key_id=local_cfg['AWS_ACCESS_KEY_ID'],
                aws_secret_access_key=local_cfg['AWS_SECRET_ACCESS_KEY'],
                aws_session_token=local_cfg['AWS_SESSION_TOKEN'])
        else:
            self.s3_client = boto3.client(
                's3', region_name='ap-southeast-1',
                aws_access_key_id=local_cfg['AWS_ACCESS_KEY_ID'],
                aws_secret_access_key=local_cfg['AWS_SECRET_ACCESS_KEY'])

    def push_bag(self, args):
        """
        Push bag to S3 and bag metadata to Cognicept cloud
        """
        upload_status = False
        if args.bag == 'latest':
            bag_file_path = self.get_latest_bag(args)
            print(
                bcolors.OKBLUE +
                "Uploading latest bag file to S3: " +
                bcolors.ENDC + bag_file_path)
            upload_status = self.upload_s3(bag_file_path, args.config)
        elif args.bag.endswith('.bag'):
            bag_file_path = os.path.expanduser(args.path+"bags/"+args.bag)
            if(self.check_bag_exists(bag_file_path)):
                print(
                    bcolors.OKBLUE +
                    "Uploading bag file to S3: " +
                    bcolors.ENDC + bag_file_path)
                upload_status = self.upload_s3(bag_file_path, args.config)
            else:
                print(
                    bcolors.FAIL +
                    "Specified bag file not found in expected location: " +
                    bcolors.ENDC + bag_file_path)
        else:
            print(
                bcolors.FAIL + "Only .bag files can be uploaded." + bcolors.ENDC)

        if upload_status:
            print(bcolors.OKGREEN + "\nBag file uploaded." + bcolors.ENDC)
            self.post_bag_metadata(args, bag_file_path)
        else:
            pass

    def get_latest_bag(self, args):
        """
        Get path of the latest bag file
        """
        bag_path = os.path.expanduser(args.path+"bags")
        bag_list = glob.glob(bag_path+"/*.bag")
        if bag_list:
            latest_file = max(bag_list, key=os.path.getctime)
            return latest_file
        else:
            print(
                bcolors.FAIL + "No bags found in the expected path." + bcolors.ENDC)
            raise(SystemExit)

    def check_bag_exists(self, bag_file_path):
        """
        Check if rosbag exists
        """
        if (os.path.exists(bag_file_path)):
            return True
        else:
            return False

    def upload_s3(self, bag_file_path, config_obj):
        """
        Upload rosbag to S3
        """
        # Multipart threaded approach for larger files
        config = TransferConfig(multipart_threshold=1024*25, max_concurrency=10,
                                multipart_chunksize=1024*25, use_threads=True)
        num_retries = 3

        for trial in range(num_retries):
            try:
                # Upload file
                file_name = ntpath.basename(bag_file_path)
                self.s3_client.upload_file(bag_file_path, self.bucket_name, file_name,
                                           Config=config, Callback=ProgressPercentage(bag_file_path))
                # Return true for successful upload
                return True
            except boto3.exceptions.S3UploadFailedError:
                # On failure, retry
                print('Attempt #' + str(trial+1) + bcolors.FAIL +
                      " FAILED" + bcolors.ENDC + "\033[K")
                # Wait for 1 second before retrying
                time.sleep(1.0)
        # If the loop is completed, upload failed, so return false
        return False

    def post_bag_metadata(self, args, bag_file_path):
        """
        Post bag metadata to Cognicept cloud
        """
        payload = {
            "robot_id": args.config.get_field("ROBOT_CODE"),
            "bagfile_name": ntpath.basename(bag_file_path),
            "bagfile_url": ntpath.basename(bag_file_path),
            "bagfile_size": str(os.path.getsize(bag_file_path)),
            "uploaded_at": datetime.datetime.now().isoformat()
        }

        # retrieve cognicept access key
        try:
            token = args.config.get_field("COGNICEPT_ACCESS_KEY")
        except KeyError:
            print(bcolors.FAIL +
                  'COGNICEPT_ACCESS_KEY variable undefined in configuration. Could not update Cognicept cloud with bag metadata.' +
                  bcolors.ENDC)
            raise(SystemExit)

        headers = {"Authorization": "Basic " + token}

        try:
            resp = requests.post(args.config.get_cognicept_api_uri(
            ) + "bagfile", headers=headers, json=payload, timeout=5)

            if resp.status_code != 200:
                print(
                    'Cognicept REST API error: ' +
                    args.config.get_cognicept_api_uri() +
                    ' responded with ' + str(resp.status_code) +
                    '\n' + resp.json()['Message'])
                return False
            else:
                print(
                    bcolors.OKGREEN +
                    "Metadata posted to Cognicept Cloud." +
                    bcolors.ENDC)
        except requests.exceptions.Timeout:
            print("Cognicept REST API error: time out.")
            return False
        except requests.exceptions.TooManyRedirects:
            print("Cognicept REST API error: Wrong endpoint.")
            return False
        except Exception as ex:
            print("Cognicept REST API error, " + str(ex))
            return False



# Copyright 2020 Cognicept Systems
# Author: Jakub Tomasek (jakub@cognicept.systems)
# --> Main executable file to handle input arguments

import argparse
import os
import sys
import pkg_resources

from cogniceptshell.configuration import Configuration
from cogniceptshell.agent_life_cycle import AgentLifeCycle
from cogniceptshell.rosbag_record import RosbagRecord
from cogniceptshell.pusher import Pusher


def main():

    # Create the parser
    parser = argparse.ArgumentParser(
        description='Shell utility to configure Cognicept tools.')

    parser.add_argument('--version', action='version',
                        version=pkg_resources.require("cognicept-shell")[0].version)

    subparsers = parser.add_subparsers(help='', title="Commands")

    parser_config = subparsers.add_parser(
        'config', help='Configure Cognicept tools')
    parser_status = subparsers.add_parser(
        'status', help='Get status of Cognicept agents')
    parser_lastevent = subparsers.add_parser(
        'lastevent', help='Display last event log reported by Cognicept agent')
    parser_update = subparsers.add_parser(
        'update', help='Update Cognicept tools')
    parser_keyrotate = subparsers.add_parser(
        'keyrotate', help='Rotate Cognicept cloud keys')
    parser_restart = subparsers.add_parser(
        'restart', help='Restart Cognicept agents')
    parser_start = subparsers.add_parser(
        'start', help='Start Cognicept agents')
    parser_stop = subparsers.add_parser('stop', help='Stops Cognicept agents')
    parser_orbitty = subparsers.add_parser('orbitty', help='Run Orbitty')
    parser_record = subparsers.add_parser(
        'record', help='Manages rosbag recording')
    parser_push = subparsers.add_parser(
        'push', help='Pushes stuff to Cognicept cloud')

    local_cfg = Configuration()
    DEFAULT_PATH = "~/.cognicept/"
    parser_config.set_defaults(func=local_cfg.configure)
    parser_config.add_argument(
        '--path', help='Cognicept configuration directory (default: `' + DEFAULT_PATH + '`)', default=DEFAULT_PATH)
    parser_config.add_argument(
        '--add',  help='Add new environmental variable in config file', action='store_true')
    parser_config.add_argument(
        '--ssh',  help='Configure ssh access during remote intervention', action='store_true')
    parser_config.add_argument(
        '--read',  help='Prints Cognicept configuration', action='store_true')

    parser_keyrotate.set_defaults(func=local_cfg.cognicept_key_rotate)
    parser_keyrotate.add_argument(
        '--path', help='Cognicept configuration directory (default: `' + DEFAULT_PATH + '`)', default=DEFAULT_PATH)

    agent_lifetime = AgentLifeCycle()
    parser_status.set_defaults(func=agent_lifetime.status)
    parser_status.add_argument(
        '--path', help='Cognicept configuration directory (default: `' + DEFAULT_PATH + '`)', default=DEFAULT_PATH)

    parser_lastevent.set_defaults(func=agent_lifetime.get_last_event)
    parser_lastevent.add_argument(
        '--path', help='Cognicept configuration directory (default: `' + DEFAULT_PATH + '`)', default=DEFAULT_PATH)

    parser_restart.set_defaults(func=agent_lifetime.restart)
    parser_restart.add_argument(
        '--path', help='Cognicept configuration directory (default: `' + DEFAULT_PATH + '`)', default=DEFAULT_PATH)
    parser_restart.add_argument(
        'list', help='List of agents to restart', metavar='list', type=str, nargs='*')
    parser_restart.add_argument(
        '--datadog', help='Restart only datadog agent', action='store_true')
    parser_restart.add_argument(
        '--agents', help='Restart only cognicept agents', action='store_true')

    parser_start.set_defaults(func=agent_lifetime.start)
    parser_start.add_argument(
        '--path', help='Cognicept configuration directory (default: `' + DEFAULT_PATH + '`)', default=DEFAULT_PATH)
    parser_start.add_argument(
        'list', help='List of agents to start', metavar='list', type=str, nargs='*')
    parser_start.add_argument(
        '--datadog', help='Restart only datadog agent', action='store_true')
    parser_start.add_argument(
        '--agents', help='Restart only cognicept agents', action='store_true')

    parser_stop.set_defaults(func=agent_lifetime.stop)
    parser_stop.add_argument(
        '--path', help='Cognicept configuration directory (default: `' + DEFAULT_PATH + '`)', default=DEFAULT_PATH)
    parser_stop.add_argument(
        'list', help='List of agents to stop', metavar='list', type=str, nargs='*')
    parser_stop.add_argument(
        '--datadog', help='Stop only datadog agent', action='store_true')
    parser_stop.add_argument(
        '--agents', help='Stop only cognicept agents', action='store_true')

    parser_update.set_defaults(func=agent_lifetime.update)
    parser_update.add_argument(
        '--path', help='Cognicept configuration directory (default: `' + DEFAULT_PATH + '`)', default=DEFAULT_PATH)
    parser_update.add_argument(
        '--reset', help='Triggers new login before update', action='store_true')

    parser_orbitty.set_defaults(func=agent_lifetime.run_orbitty)
    parser_orbitty.add_argument(
        '--path', help='Cognicept configuration directory (default: `' + DEFAULT_PATH + '`)', default=DEFAULT_PATH)

    record_session = RosbagRecord()
    parser_record.set_defaults(func=record_session.record)
    parser_record.add_argument(
        '--path', help='Cognicept configuration directory (default: `' + DEFAULT_PATH + '`)', default=DEFAULT_PATH)
    parser_record.add_argument(
        '--start', help='Start rosbag recording session. Provide topics to record separated by spaces. e.g. cognicept record --start rosout odom', nargs='+')
    parser_record.add_argument(
        '--stop', help='Stop rosbag recording session', action='store_true')
    parser_record.add_argument(
        '--pause', help='Pause rosbag recording session', action='store_true')
    parser_record.add_argument(
        '--resume', help='Resume rosbag recording session', action='store_true')
    parser_record.add_argument(
        '--status', help='Get current recording session status', action='store_true')

    pusher_instance = Pusher()
    parser_push.set_defaults(func=pusher_instance.push)
    parser_push.add_argument(
        '--path', help='Cognicept configuration directory (default: `' + DEFAULT_PATH + '`)', default=DEFAULT_PATH)
    parser_push.add_argument(
        '--bag', help='Pushes rosbag to the configured S3 bucket. Set to BAG to latest to upload latest bag')
    # Parse the arguments
    args = parser.parse_args()

    if("path" not in args):
        parser.print_help()
    else:
        local_cfg.load_config(args.path)
        agent_lifetime.configure_containers(local_cfg)
        args.config = local_cfg

        if(hasattr(args, 'func')):
            args.func(args)


if __name__ == "__main__":
    main()




import pytest
import requests
from mock import patch
import os

from cogniceptshell.configuration import Configuration


class SuccessAwsCredentials(object):
    def __init__(self):
        self.status_code = 200

    def json(self):
        return {"AccessKeyId": os.getenv('AWS_ACCESS_KEY_ID', ""), "SecretAccessKey": os.getenv('AWS_SECRET_ACCESS_KEY', ""), "SessionToken": ""}


class NotFoundMockResponse(object):
    def __init__(self):
        self.status_code = 404


def mock_aws_endpoint(*args, **kwargs):
    if(args[0] == "https://test.cognicept.systems/api/agent/v1/aws/assume_role"):
        return SuccessAwsCredentials()
    else:
        return NotFoundMockResponse()

def setup_file(tmpdir):
    p = tmpdir.join("runtime.env")
    p.write("COG_AGENT_CONTAINERS=container1;container2\nCOG_AGENT_IMAGES=image1;image2\nCOGNICEPT_API_URI=https://test.cognicept.systems/api/agent/v1/aws/assume_role\nCOGNICEPT_ACCESS_KEY=CORRECT-KEY")


def setup_wrong_uri_file(tmpdir):
    p = tmpdir.join("runtime.env")
    p.write(
        "COGNICEPT_API_URI=https://www.wronguri.blame\nCOGNICEPT_ACCESS_KEY=INCORRECT-KEY")


def setup_correct_uri_file(tmpdir):
    p = tmpdir.join("runtime.env")
    p.write(
        "COGNICEPT_API_URI=https://test.cognicept.systems/api/agent/v1/\nCOGNICEPT_ACCESS_KEY=CORRECT-KEY")

def test_yes_input():
    object = Configuration()
    assert (object._interpret_bool_input("Y") == True)

def test_no_input():
    object = Configuration()
    assert (object._interpret_bool_input("n") == False)

def test_other_input():
    object = Configuration()
    assert (object._interpret_bool_input("g") == None)
    assert (object._interpret_bool_input("%") == None)
    assert (object._interpret_bool_input("1") == None)
    assert (object._interpret_bool_input("akjflakjewr4f56f74ew@!!@$@!$") == None)

def test_is_ssh_disabled(tmpdir):
    setup_file(tmpdir)
    object = Configuration()
    object.load_config(str(tmpdir) + "/")
    assert (object.is_ssh_enabled() == False)    
    object.config["COG_ENABLE_SSH_KEY_AUTH"] = False
    assert (object.is_ssh_enabled() == False)

def test_is_ssh_disabled(tmpdir):
    setup_file(tmpdir)
    object = Configuration()
    object.load_config(str(tmpdir) + "/")
    object.config["COG_ENABLE_SSH_KEY_AUTH"] = True
    assert (object.is_ssh_enabled() == True)


def test_incorrect_cognicept_key_fetch_aws_keys(tmpdir, capsys, monkeypatch):
    setup_wrong_uri_file(tmpdir)
    
    monkeypatch.setattr(requests, "get", mock_aws_endpoint)
    object = Configuration()
    object.load_config(str(tmpdir) + "/")
    try:
        result = object.fetch_aws_keys()
        assert(result == False)
    except:
        pytest.fail("Incorrect Cognicept API URI gave exception", pytrace=True)


def test_correct_cognicept_key_fetch_aws_keys(tmpdir, capsys, monkeypatch):
    setup_correct_uri_file(tmpdir)

    monkeypatch.setattr(requests, "get", mock_aws_endpoint)
    object = Configuration()
    object.load_config(str(tmpdir) + "/")
    try:
        result = object.fetch_aws_keys()
        assert(result == SuccessAwsCredentials().json())
    except:
        pytest.fail("Correct Cognicept API URI gave exception", pytrace=True)
    
